#
# Z. Dodds
# Ben Tribelhorn
#
# An implementation of MCL.
#

START_ON_ROBOT = False
USE_OUTOFBOUNDS = True

USE_RANGE_VALUES = True
USE_ONE_RANGE_VALUE = False
USE_ONE_RANGE_VALUE_NG = False

NOISE_PERCENTAGE = 0.25   # 0.25 means 25% not 0.25% ! 
ANOISE_PERCENTAGE = 0.10   # 0.25 means 25% not 0.25% ! 

# python modules with things we need...
import time                # fun with clocks!
import math                # all the familiar functions
import random              # maybe we'll use it, maybe we won't


# these are constants for accessing the simulated range values
BaRi = 0    # the index of the Back Right sensor reading
BaLe = 1    # the index of the Back Left sensor reading
FrLeSi = 2  # the index of the Front Left side-looking sensor reading
FrLe = 3    # the index of the Front Left front-looking sensor reading
Fr   = 4    # the index of the Front sensor reading
FrRi = 5    # the index of the Front Right front-looking sensor reading
FrRiSi = 6  # the index of the Front Right side-looking sensor reading

def frontThree(R):
    return R[3:6]

def angleCompare( thd1, thd2, window ):
    # make sure the angles are between 0.0 and 360.0
    while thd1 > 360.0: thd1 -= 360.0
    while thd2 > 360.0: thd2 -= 360.0
    while thd1 <   0.0: thd1 += 360.0
    while thd2 <   0.0: thd2 += 360.0
    # figure out their relative positions on the unit circle
    smaller = min(thd1,thd2)
    larger  = max(thd1,thd2)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (360.0-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway) <= window


    
    
# PL is the particle list
# each particle is a list [x,y,thr,p], where 
# p represents a relative probability (need not sum to 1)
def randomParticleStart(x=0.0, y=0.0, thr=0.0, np=100):
    NUM_PARTICLES = np
    PL = []
    if START_ON_ROBOT:
        xrange = 0
        yrange = 0
        thrrange = 0
    else:
        xrange = 90
        yrange = 60
        thrrange = math.pi
        x = 0.0
        y = 0.0
        thr = math.pi

    for i in range(0,NUM_PARTICLES):
        PL.append( [ random.uniform(x-xrange,x+xrange), \
                     random.uniform(y-yrange,y+yrange), \
                     random.uniform(thr-thrrange,thr+thrrange), \
                     1.0 ] )
    return PL


def computeSmallestAngleDifference( fromThr, toThr ):
    # make sure the angles are between 0.0 and 360.0
    twopi = 2*math.pi
    while fromThr > twopi: fromThr -= twopi
    while toThr > twopi: toThr -= twopi
    while fromThr <  0.0: fromThr += twopi
    while toThr <  0.0: toThr += twopi
    # figure out their relative positions on the unit circle
    smaller = min(fromThr, toThr)
    larger  = max(fromThr, toThr)
    # figure out the two possible differences
    oneway = larger - smaller
    otherway = (twopi-larger) + smaller
    # is the difference within the window?
    return min(oneway,otherway)

def computeSmallestSignedAngleDifference( fromThr, toThr ):
    """ returns the signed value (or smaller absolute value)
        that needs to be added to fromThr
        to yield an engle equivalent to toThr
    """
    posDiffr = computeSmallestAngleDifference( fromThr, toThr )
    # but we need to sign this to go from fromThr to toThr
    add = computeSmallestAngleDifference( fromThr + posDiffr, toThr )
    sub = computeSmallestAngleDifference( fromThr - posDiffr, toThr )
    if add < sub:
        return posDiffr
    else:
        return -posDiffr

def computeTotalRobotMotion( oldLoc, newLoc):
    oldx, oldy, oldthr = oldLoc
    x, y, thr = newLoc
    # we return thr1, dist, thr2
    # where
    # thr1 is the smallest angle in radians from oldThr
    #         to the direction of the vector (x,y)-(oldx,oldy)
    # dist is the distance between (x,y) and (oldx, oldy)
    # thr2 is the smallest angle in radians from the direction
    #         of the vector mentioned above to thr
    
    posdist = math.sqrt( (x-oldx)**2 + (y-oldy)**2 )
    # if the center hasn't moved (or only a tiny amount), we don't want
    # to call atan2, because it's not defined or is unstable numerically
    if posdist < 0.1:  # 1mm since everything is in centimeters (or should be)
        # we just have one angle difference to find
        thr1 = computeSmallestSignedAngleDifference( oldthr, thr )
        return (thr1, 0.0, 0.0)
        
    # if we did move a bit, then we use the full turn-move-turn model
    motion_thrR = math.atan2(oldy-y, oldx-x)
    motion_thrF = math.atan2(y-oldy, x-oldx)
    
    if abs(motion_thrR) < abs(motion_thrF):
        motion_thr = motion_thrR
        dist = -math.sqrt( (x-oldx)**2 + (y-oldy)**2 )
    else:
        motion_thr = motion_thrF
        dist = math.sqrt( (x-oldx)**2 + (y-oldy)**2 )
    
    thr1 = computeSmallestSignedAngleDifference( oldthr, motion_thr )
    thr2 = computeSmallestSignedAngleDifference( motion_thr, thr )
    #print 'th1,  dist,  th2 ', math.degrees(thr1), dist, math.degrees(thr2)
    return (thr1, dist, thr2)
    

def addNoise( thr1, dist, thr2 ):
    # return a triple: the same nominal values, with noise
    #
    NOISE_TYPE = 'unif_perc'
    #
    if NOISE_TYPE == 'none':
        return (thr1,dist,thr2)
        
    elif NOISE_TYPE == 'unif_perc':  # uniform noise as a percent of value
        low = min(-ANOISE_PERCENTAGE*thr1, +ANOISE_PERCENTAGE*thr1)
        high = max(-ANOISE_PERCENTAGE*thr1, +ANOISE_PERCENTAGE*thr1)
        thr1dn = random.uniform(low,high)  # delta noise
        
        low = min(-NOISE_PERCENTAGE*dist, +NOISE_PERCENTAGE*dist)
        high = max(-NOISE_PERCENTAGE*dist, +NOISE_PERCENTAGE*dist)
        distdn = random.uniform(low,high)
        
        low = min(-ANOISE_PERCENTAGE*thr2, +ANOISE_PERCENTAGE*thr2)
        high = max(-ANOISE_PERCENTAGE*thr2, +ANOISE_PERCENTAGE*thr2)
        thr2dn = random.uniform(low,high)
        
        return (thr1 + thr1dn,dist + distdn,thr2 + thr2dn)
        
    elif NOISE_TYPE == 'uniform':  # real uniform noise - not much sense here
        pass
    elif NOISE_TYPE == 'tri_perc':  # real uniform noise - not much sense here
        pass
    elif NOISE_TYPE == 'gauss_perc':  # real uniform noise - not much sense here
        pass
        
    return ('should','not','be here')  # we shouldn't get here...
    

def moveParticles( PL, thr1, dist, thr2, simmap ):
    for p in PL:
        # add noise to the values... to find thr1n, distn, thr2n
        thr1n, distn, thr2n = addNoise( thr1, dist, thr2 )
        #print 'th1,  dist,  th2 ', math.degrees(thr1), dist, math.degrees(thr2)
        #print 'th1n, distn, th2n', math.degrees(thr1n), dist, math.degrees(thr2n)
          
        #check for collisions with walls here
        dx = distn*math.cos(p[2] + thr1n)
        dy = distn*math.sin(p[2] + thr1n)
        
        # interesting... this is an update to the probability that
        # happens here in moveParticles... but it seems reasonable
        # and it's all in MCL
        if simmap.pathIntersectsWall(p[0], p[1], p[0]+dx, p[1]+dy):
            p[3] *= 0.001
        
        # now make the movement official...
        p[0] += dx
        p[1] += dy
        p[2] += (thr1n + thr2n)  # add the angles
        while p[2] > 2*math.pi:
            p[2] -= 2*math.pi
        while p[2] < 0:
            p[2] += 2*math.pi
        
    return PL

def resampleParticles( PL ):
    # what was the clever way to do this again?
    # we want to get l particles out, right...
    newPL = []
    # need the sum of the "probabilities"
    sum = 0.0
    for part in PL: sum += part[3]
    # that's 100%, now start at 0.5
    deltaThresh = sum/len(PL) # amount to raise curThresh by each time
    curThresh = deltaThresh/2.0 # seems like a safe starting point
    curPartSum = 0.0
    curPartIndex = 0
    # two nested while loops
    while curPartIndex < len(PL):  # particle loop
        curPartSum += PL[curPartIndex][3]  # add next particle's weight
        while curPartSum > curThresh:  # sampling/probability loop 
            newPL.append( PL[curPartIndex][:] ) # new copy
            newPL[-1][3] =  1.0
            curThresh += deltaThresh  # invariant: append1 for each! curThresh
        curPartIndex += 1  # next particle
    return newPL

def rangeWeight( r1, r2 ):
    # returns the update weight for these two ranges
    # if equal == 1
    # if outside 100% error == 0.1
    MIN_WEIGHT = 0.1
    # linear in between
    rmin = min( r1, r2 )
    rmax = max( r1, r2 )
    if rmin <= 0.0 or rmax > 2.0*rmin: return MIN_WEIGHT
    # percError should be between 0.0 and 1.0
    percError = (float(rmax)/float(rmin)) - 1.0
    # map to the range 1.0 -> 0.1
    relweight = (1.0-percError) * 1.0 + percError * 0.1
    # now, if within 5 cm, 1.0
    if rmax-rmin < 5.0: return 1.0
    # if outside 50 cm, 0.1*relweight
    absweight = 1.0
    if rmax-rmin > 50.0:
        absweight = 0.1/(rmax-rmin-49.0)
    return absweight*relweight
    
def outofbounds(part, simmap):
    minx,maxx,miny,maxy = simmap.getBounds()
    return part[0] < minx or part[0] > maxx or part[1] < miny or part[1] > maxy

def updateParticles( PL, simmap, rCollisions, sensors=[False, False, False]):
    """left= front= right = False
    if len(rCollisions) > 0:
        #I'm pretty confident about these numbers (for simulated robot)
        if rCollisions[0] < -0.2:
            right = True
        elif rCollisions[0] > 0.2:
            left = True
        else:
            front = True
        #left= front= right = True"""
    left, front, right = sensors

    for part in PL:
        #print "prob is ", part[3]
        if USE_OUTOFBOUNDS:
            if outofbounds(part, simmap):
                part[3] *= 0.0001
        
        if left or right: #or front
            collisions, dists = simmap.findparticlecollisions(part[0],part[1],part[2], radius=20)   #increased to save
            if len(collisions) == 0:
                if len(dists) > 0:  #is out there between 20 and 33
                    part[3] *= 0.01
                    #part[3] *= 1.01 - abs(16.5 - dists[0])/16.5#0.1
                else:   #farther than twice the radius of the roomba
                    part[3] *= 0.001
            else:
                if abs(16.5 - dists[0]) > 2.5:    #if it's from 14-19 we're at a flat peak else:
                    part[3] *= 1.01 - abs(16.5 - dists[0])/16.5
        else:
            collisions, dists = simmap.findparticlecollisions(part[0],part[1],part[2], radius=14)   #reduced radius to save particle
            if len(collisions) > 0:
                part[3] *= 0.001
        #print "3prob is ", part[3]
        if len(collisions) > 0:
            if left and not front:
                if collisions[0] > 1.8:
                    print "not a left collision ", part[2], collisions[0]
                    part[3] *= 0.1
                #if computeSmallestAngleDifference(part[2], collisions[0]) > math.pi/6.:
                 #   print "left hit is (too) dif : ", part[2], collisions[0]
                  #  part[3] *= 0.1
            elif right and not front:
                if collisions[0] < -1.8:
                    print "not a right collision ", part[2], collisions[0]
                    part[3] *= 0.1
            elif front:
                if math.fabs(collisions[0]) > 0.8:
                    print "not a front collision ", part[2], collisions[0]
                    part[3] *= 0.1
        #print "final prob ", part[3]

    """
    (x,y,thr,(R,stalled),timestamp,(rx,ry,rthr)) = ALL_SENSORS
    for part in PL:
        # we need to update its probability
        # first, we get its sensor state
        RP, numhitsP = robot.GetAllDataFrom(part[0],part[1],part[2]) 
        #
        # compare stalled and numhits (stalled is False!)
        if USE_OUTOFBOUNDS:
            if numhitsP % 2 == 0:
                part[3] *= 0.0001
        #
        # compare R and RP
        # first, just forward...

        if USE_ONE_RANGE_VALUE_NG and R[Fr] - RP[Fr] > 10.0:
            #for i in range(0,7):
                part[3] *= 0.001

        if USE_ONE_RANGE_VALUE:
            #for i in range(0,7):
                part[3] *= rangeWeight( RP[Fr], R[Fr] )

        if USE_RANGE_VALUES:
            for i in range(0,7):
                part[3] *= rangeWeight( RP[i], R[i] )
        #BaRi = 0    # the index of the Back Right sensor reading
        #BaLe = 1    # the index of the Back Left sensor reading
        #FrLeSi = 2  # the index of the Front Left side-looking sensor reading
        #FrLe = 3    # the index of the Front Left front-looking sensor reading
        #Fr   = 4    # the index of the Front sensor reading
        #FrRi = 5    # the index of the Front Right front-looking sensor reading
        #FrRiSi = 6  # the index of the Front Right side-looking sensor reading
    """
    # now resample the particles
    PL = resampleParticles( PL )
    return PL


class MCL:
    def __init__(self, mapobject, np=100):
        self.simmap = mapobject
        self.oldpos = [0, 0, 0]
        self.NUM_PARTS = np
        
        self.PL = randomParticleStart(np=self.NUM_PARTS)
    
    def pose(self, i):
        return self.PL[i][0:3]
    
    def updateMCL(self, collisions, newpos, sensors=[False, False, False]):
        #MCL
        thr1, dist, thr2 = computeTotalRobotMotion(self.oldpos, newpos)
        moveParticles(self.PL, thr1, dist, thr2, self.simmap)
                
        self.PL = updateParticles( self.PL, self.simmap, collisions, sensors)
        self.oldpos = newpos

    def resetMCL(self):
        self.PL = randomParticleStart(np=self.NUM_PARTS)



